<html>
<head>
   <meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">
   <meta name="Author" content="Peter Kunszt">
   <title>HTM Index C++ Reference</title>
<!--
Authors:        Peter Z. Kunszt
Created:        August 2000
-->
</head>

<body text="#111111" bgcolor="#FFFFFF" link="#333399" vlink="#663366" alink="#CC0000">

<h2><IMG SRC = "htmsphere.jpg">&nbsp;&nbsp;
<font color="#7700EE">HTM SpatialIndex Hierarchy</font>
</h2>
<img SRC="rule.gif" >
<br><br>

