//# SpatialDoc.h
//#
//# Just for documentations' sake.
//#
//# Author:		Peter Z. Kunszt
//#	
//# Creation:		October 19, 1999
//#
//# (c) Copyright The Johns Hopkins University 1999
//# All Rights Reserved
//#
//# The software and information contained herein are proprietary to The
//# Johns Hopkins University, Copyright 1995, 1996. This software is furnished
//# pursuant to a written license agreement and may be used, copied,
//# transmitted, and stored only in accordance with the terms of such
//# license and with the inclusion of the above copyright notice.  This
//# software and information or any other copies thereof may not be
//# provided or otherwise made available to any other person.
//#
//# Modification History:

/** Types Defined by the SpatialIndex Package. (if not already defined)


<pre>
 bool           Boolean type, and the constants 'true' and 'false'

 int8           Integer types up to 64 bits
 int16
 int32
 int64

 uint8          Unsigned integer types
 uint16
 uint32
 uint64

 float32        Float and double precision.
 float64
</pre>
*/
#define HTMTYPES

